var util = require('util');
var path = require('path');
var wizzi = require('../wizzi');
var log = require('../util/log')(module, { tofile: true });
var node = require('../util/node');
var verify = wizzi.verify;
var ProductionContext = require("../production/context");
var ModelProvider = require('./provider');
var LoadContext = require('./loadContext');
var IttfModel = require('./ittfModel');
var includer = require('./includer');
var mixer = require('./mixer');
var appender = require('./appender');
var evaluator = require('./evaluator2');

var md = module.exports = {};

/*
    The Big JOB of an IttfLoad
    manages
        lexer
        parser
        includer
        mixer
        appender
        evaluator
        normalize context objects
        instantiate the ittfLoadState
        instantiate the modelProvider
*/

md.load = function (ittfDocumentUri, requestContext, callback) {

    if (typeof callback !== 'function') {
        throw new Error("Ittf.load. A callback is required. Filepath: " + ittfDocumentUri);
    }

    requestContext = normalizeRequestContext(requestContext);
    var ittfLoadState = {
        requestContext: requestContext,
        provider: ModelProvider.createFromFilepath(ittfDocumentUri, requestContext),
        callChain: [],
        checkForLoops: function () {
            for (var i = 0; i < this.callChain.length; i++) {
                if (this.callChain[i].caller == this.callChain[i].callee) return true;
                for (var j = 0; j < this.callChain.length; j++) {
                    if ((this.callChain[j].caller == this.callChain[i].callee) &&
                        (this.callChain[j].callee == this.callChain[i].caller)) return true;
                }
            }
            return false; 
        }
    }

    ittfLoadState.provider.get(
        {
            from: 'store',
            ittfDocumentUri: ittfLoadState.provider.getPrimaryModelFilepath()
                // clumsy but what to do ???
                // it is so for the primaryModel
                // for mixed models (see ./mixer) parameters are different
        },
        function (err, ittfModel) {
            // log.info('ittf.load.got.ittfModel', ittfModel);
            // console.log("ittf.load.includes");
            if (err) { return callback(err); }
            ittfLoadState.callChain.push({
                caller: null,
                callee: ittfModel.uri
            });
            includer(ittfModel, ittfLoadState, function (err, includeResult) {
                if (err) { return callback(err); }
                mixer(includeResult, ittfLoadState, function (err, mixResult) {
                    if (err) { return callback(err); }
                    requestContext.productionContext.addMixedIttfModel(
                        ittfDocumentUri,
                        mixResult
                        );
                    // console.log('mixResult ' + mixResult.uri);
                    // node.dump(mixResult.nodes);
                    try {
                        // sync calls
                        ittfModel = appender(mixResult);
                        ittfModel = evaluator(ittfModel, requestContext);
                        ittfModel.loadContext = ittfLoadState.provider.loadContext;
                        requestContext.productionContext.addEvaluatedIttfModel(
                            ittfDocumentUri,
                            ittfModel
                            );
                    } catch (ex) {
                        var error = new wizzi.errors.IttfLoadError(
                            'Exception: ' + ex.message,
                            ittfModel.uri,
                            ex
                            )
                        wizzi.fail.warn(error.message);
                        return callback(error);
                    }
                    // console.log('+++ ittfModel', ittfModel.nodes)
                    for (var key in ittfModel) console.log('ittfModel prop: ' + key);
                    callback(null, ittfModel);
                });
            });
        }
    );
}

md.loadFromString = function (content, requestContext, callback) {
    // console.log('ittf.loadFromString.content', content);
    var ittfLoadState = {
        requestContext: normalizeRequestContext(requestContext),
        provider: ModelProvider.createFromFilepath(ittfDocumentUri, requestContext)
    }
    var loadContext = new LoadContext();
    var ittfModel = new IttfModel(loadContext, requestContext.modelContext);
    ittfModel.load(content, function (err, ittfModel) {
        if (err) { return callback(err); }
        loadContext.addIttfModel('text://', null, ittfModel, {});
        includer(ittfModel, ittfLoadState, function (err, includeResult) {
            if (err) { return callback(err); }
            mixer(includeResult, ittfLoadState, function (err, mixResult) {
                if (err) { return callback(err); }
                ittfModel = appender(mixResult);
                ittfModel = evaluator(ittfModel, requestContext);
                // console.log('ittf.loadFromString.ittfModel', ittfModel);
                callback(null, ittfModel);
            });
        });
    });
}
/*
    An IttfLoad execution contains 3 context object
        modelContext       IttfModel evaluation context (see evaluator.js)
        artifactContext    ArtifactGeneration context 
                           not used during IttfLoad ??? why is it here ???
        productionContext  A production context that may span more then one IttfLoad
*/

function normalizeRequestContext(context) {
    if (verify.isObject(context)) {
        if (verify.isObject(context.productionContext)) {
            if (verify.isObject(context.modelContext)) {
                if (verify.isObject(context.artifactContext)) {
                    return context;
                } else {
                    context.artifactContext = context.modelContext;
                    return context;
                }
            } else {
                context.modelContext = {};
                context.artifactContext = {};
                return context;
            }
        } else {
            // console.log('ProductionContext', ProductionContext)
            return {
                modelContext: context,
                artifactContext: context,
                productionContext: new ProductionContext()
            }
        }
    } else {
        return {
            modelContext: {},
            artifactContext: {},
            productionContext: new ProductionContext()
        }
    }
}

