var util = require('util');
var path = require('path');
var url = require('url');
var wizzi = require('../wizzi');
var log = require('../util/log')(module, { tofile: true });
var verify = wizzi.verify;
var ModelStore = require('./store').ModelStore;
var IttfModel = require("./ittfModel");
var LoadContext = require('./loadContext');

/*
    // is created for each IttfLoading
    modelProvider : {
        schema: String,
        ittfSource: {
            path: uri,
            storeKind: String ['repo'|'filesystem'], 
            storeDB: String 
        },
        store: Object, // new ModelStore(ittfSource)
        loadContext: Object, // new LoadContext(),
        productionContext: Object, // context.productionContext
        _modelsCache: Array // of IttfModel
    }
*/

var md = module.exports = {};

/*
    params 
        options {
            schema: String
            ittfSource: Object
            store: Object
            loadContext: Object
            productionContext: Object
        }
*/
md.modelProvider = function (options) {
    this.schema = options.schema; // A provider instance manages ittf documents of one schema
    this.store = options.store;
    this.loadContext = options.loadContext;
    if (!options.productionContext) {
        // Must exists. Brutal error.
        throw new Error("wizzi.core.lib.ittf.provider.options. Missing productionContext. options: " +
            util.inspect(options));
    }
    this.productionContext = options.productionContext;
    this.ittfSource = options.ittfSource;
    this._modelsCache = {};
}

/*
    params 
        options {
            from: String [store|string]
            ittfDocumentUri: String 
            include: Boolean
            includerModelKey: String
            basedir: String
            relpath: String
        }
    called from
        ./ittf.load()
        ./includer()
        ./mixer.mix()
*/
md.modelProvider.prototype.get = function (options, callback) {
    if (options.from === 'store') {
        return this.getFromStore(options, callback);
    } else {
        return this.getFromString(options, callback);
    }
}

md.modelProvider.prototype.getPrimaryModelFilepath = function () {
    return this.ittfSource.path;
}

md.modelProvider.prototype.getFromStore = function (options, callback) {
    var loadContext = this.loadContext;
    var productionContext = this.productionContext;
    var self = this;
    this.resolvePath(options, function (err, uri) {
        if (err) { return callback(err); }
        // console.log('modelProvider.getFromStore.uri', uri, self._modelsCache);
        var retModel = null;
        var ittfModel = self._modelsCache[uri];
        if (ittfModel) {
            retModel = ittfModel.clone();
            productionContext.addIttfDocument(uri, ittfModel.inputContent);
            var ittfDocumentData = loadContext.addIttfModel(uri, self.schema, retModel, options);
            retModel.sourceKey = ittfDocumentData.sourceKey;
            retModel.modelKey = ittfDocumentData.modelKey;
            retModel.$schema = self.schema;
            return callback(null, retModel);
        } else {
            self.store.getModelContent(uri, function (err, modelContent) {
                if (err) { return callback(err); }
                productionContext.addIttfDocument(uri, modelContent);
                loadContext.addIttfDocument(uri, modelContent);
                var ittfDocumentData = loadContext.addIttfModel(uri, self.schema, null, options);
                ittfModel = new IttfModel(
                    uri, loadContext
                );
                try
                {
                    // sync call
                    ittfModel.load(modelContent, ittfDocumentData);
                } catch (ex) {
                    var error = new wizzi.errors.IttfLoadError(ex.message, uri, ex);
                    return callback(error);
                }
                
                self._modelsCache[uri] = ittfModel;
                retModel = ittfModel.clone();
                retModel.sourceKey = ittfDocumentData.sourceKey;
                retModel.modelKey = ittfDocumentData.modelKey;
                retModel.$schema = self.schema;
                if (!options.include) {
                    // console.log('ittfDocumentData.setittfModel', ittfDocumentData.setittfModel);
                    ittfDocumentData.ittfModel = retModel;
                }
                // console.log('modelProvider.retModel.$schema', retModel.$schema);
                // throw new Error('stop');
                return callback(null, retModel);
            });
        }
    });
}

md.modelProvider.prototype.resolvePath = function (options, callback) {
    var ittfDocumentUri = options.ittfDocumentUri,
        relpath = options.relpath;

    if (typeof ittfDocumentUri === 'undefined') {
        if (typeof relpath === 'undefined' || relpath[0] === '/') {
            throw new Error('A "relpath" options with a relative path is required to load an ITTF document, or use the "ittfDocumentUri" option');
        }
        var basedir = options.basedir;
        if (typeof basedir === 'undefined') {
            throw new Error('A "basedir" option is required to load an ITTF document using a relative path');
        }
        ittfDocumentUri = path.join(basedir, relpath);
    }

    var self = this;
    this.tryExists(ittfDocumentUri, this.schema, function (err, result) {
        // log.info('modelProvider.resolvePath.tryExists.ittfDocumentUri/schema/result', ittfDocumentUri, self.scchema, result);
        if (err) { return callback(err); }
        if (result.found) {
            // console.log('resolvePath.result.ittfDocumentUri', result.ittfDocumentUri);
            return callback(null, result.ittfDocumentUri);
        } else {
            if (typeof relpath === 'undefined') {
                var error = new wizzi.errors.NotFoundError('IttfDocument', ittfDocumentUri, self.ittfSource.path);
                return callback(error);
            }
            self.resolvePathInTFolders(path.dirname(ittfDocumentUri), relpath, function (err, tresult) {
                if (err) { return callback(err); }
                if (tresult.found)
                    return callback(null, tresult.ittfDocumentUri);
                else {
                    var error = new wizzi.errors.NotFoundError('IttfDocument', ittfDocumentUri, self.ittfSource.path);
                    return callback(error);
                }
            });
        }
    });
}

md.modelProvider.prototype.resolvePathInTFolders = function (basePath, relPath, callback) {
    var ittfDocumentUri = path.join(basePath, 't', relPath);
    var self = this;
    this.tryExists(ittfDocumentUri, this.schema, function (err, result) {
        if (err) { return callback(err); }
        if (result.found)
            return callback(null, result);
        else {
            basePath = path.dirname(basePath);
            if (basePath.length > 3) { // TODO limit to app root
                return self.resolvePathInTFolders(basePath, relPath, callback);
            } else {
                return callback(null, { found: false });
            }
        }
    });
}

md.modelProvider.prototype.tryExists = function (test, schema, callback) {
    // log.info('modelProvider.resolvePath.tryExists.starting.test/schema', test, schema);
    var self = this;
    this.store.documentExists(test, function (err, result) {
        // console.log('tryExists', test, err, result)
        if (err) { return callback(err); }
        // console.log('tryExists 2', result)
        if (result) {
            // console.log('tryExists ok', test, err, result)
            return callback(null, { found: true, ittfDocumentUri: test });
        } else {
            // console.log('tryExists not ok', test, err, result)
            if (test.toLowerCase().substr(-5, 5) !== '.ittf') {
                if (schema) { // try implicit schema extension
                    self.store.documentExists(test + '.' + schema + '.ittf', function (err, result) {
                        if (err) { return callback(err); }
                        if (result) {
                            return callback(null, { found: true, ittfDocumentUri: test + '.' + schema + '.ittf' });
                        } else { // try ittf extension
                            self.store.documentExists(test + '.ittf', function (err, result) {
                                if (err) { return callback(err); }
                                if (result){
                                    return callback(null, { found: true, ittfDocumentUri: test + '.ittf' });
                                } else { // NOT FOUND
                                    return callback(null, { found: false });
                                }
                            });
                        }
                    });
                } else {
                    return callback(null, { found: false });
                }
            } else {
                return callback(null, { found: false });
            }
        }
    });
}

md.modelProvider.prototype.getFromString = function (options, callback) {
    throw new Error("modelProvider.getFromString not implemented");
}

// Static methods

md.createFromFilepath = function (filepath, context) {

    var fileUrl = url.parse(filepath);
    if (typeof fileUrl.protocol !== 'string') {
        throw new Error("Ittf.Provider.createFromFilepath. filepath must have a protocol (must be an absolute url) : " + filepath);
    }

    // log.info('Ittf.Provider.createFromFilepath.filepath/url', filepath, fileUrl);
    var protocol = fileUrl.protocol.substr(-1, 1) === ':'
        ? fileUrl.protocol.substr(0, fileUrl.protocol.length - 1)
        : fileUrl.protocol;
    var ittfSource = {};
    if (protocol.length === 1) {
        ittfSource.storeKind = 'filesystem';
        ittfSource.path = filepath;
    } else if (protocol === 'repo') {
        ittfSource.storeKind = 'repo';
        ittfSource.storeDB = fileUrl.host;
        ittfSource.path = fileUrl.path.substr(1);
    } else {
        throw new Error('Error Ittf.Provider.createFromFilepath. Unknown protocol: ' + protocol);
    }
    // log.info('Ittf.Provider.createFromFilepath.ittfSource', ittfSource);

    return new md.modelProvider({
        schema: inferSchema(filepath),
        ittfSource: ittfSource,
        store: new ModelStore(ittfSource),
        loadContext: new LoadContext(),
        productionContext: context.productionContext
    });

}

function inferSchema(filepath) {
    var name = path.basename(filepath).toLowerCase();
    var parts = name.split('.');
    // console.log('parts', name, parts[parts.length - 1], parts[parts.length - 2]);
    var schema = parts.length > 1 && parts[parts.length - 1] === 'ittf' ? parts[parts.length - 2] : null;
    // console.log('ittf.inferSchema', filepath, schema);
    return schema;
}